@@ -34,6 +34,7 @@
 #include "xrdp_constants.h"
 #include "xrdp_sockets.h"
 #include "chansrv_common.h"
+#include "chansrv_config.h"
 #include "list.h"
 #include "audin.h"
 
@@ -60,6 +61,7 @@ static lame_global_flags *g_lame_encoder = 0;
 
 extern int g_rdpsnd_chan_id;    /* in chansrv.c */
 extern int g_display_num;       /* in chansrv.c */
+extern struct config_chansrv *g_cfg; /* in chansrv.c */
 
 /* audio out: sound_server -> xrdp -> NeutrinoRDP */
 static struct trans *g_audio_l_trans_out = 0; /* listener */
@@ -72,7 +74,7 @@ static struct trans *g_audio_c_trans_in = 0;  /* connection */
 static int    g_training_sent_time = 0;
 static int    g_cBlockNo = 0;
 static int    g_bytes_in_stream = 0;
-FIFO   g_in_fifo;
+struct fifo  *g_in_fifo;
 int    g_bytes_in_fifo = 0;
 static int    g_time_diff = 0;
 static int    g_best_time_diff = 0;
@@ -1099,9 +1101,9 @@ process_pcm_message(int id, int size, struct stream *s)
         case 0:
             if ((g_client_does_fdk_aac || g_client_does_mp3lame) && sending_silence)
             {
-                if ((g_time3() - silence_start_time) < 1000)
+                if ((g_time3() - silence_start_time) < (int)g_cfg->msec_do_not_send)
                 {
-                    /* do not send data within 1000mS after SNDC_CLOSE is sent. to avoid stutter */
+                    /* do not send data within msec_do_not_send msec after SNDC_CLOSE is sent, to avoid stutter. setting from sesman.ini */
                     return 0;
                 }
                 sending_silence = 0;
@@ -1112,12 +1114,11 @@ process_pcm_message(int id, int size, struct stream *s)
             if ((g_client_does_fdk_aac || g_client_does_mp3lame) && sending_silence == 0)
             {
                 /* workaround for mstsc.exe. send silence data before send close */
-                int send_silence_times = g_client_does_fdk_aac ? 4 : 2;  /* This value comes by trial and error */
+                int send_silence_times = g_client_does_fdk_aac ? g_cfg->num_silent_frames_aac : g_cfg->num_silent_frames_mp3;  /* setting from sesman.ini */
                 char *buf = (char *) g_malloc(g_bbuf_size, 0);
                 if (buf != NULL)
                 {
                     int i;
-
                     silence_start_time = g_time3();
                     sending_silence = 1;
                     for (i = 0; i < send_silence_times; i++)
@@ -1259,6 +1260,15 @@ sound_sndsrvr_source_conn_in(struct trans *trans, struct trans *new_trans)
     return 0;
 }
 
+/*****************************************************************************/
+/* Item destructor for g_in_fifo
+ */
+static void
+in_fifo_item_destructor(void *item, void *closure)
+{
+    xstream_free((struct stream *)item);
+}
+
 /*****************************************************************************/
 int
 sound_init(void)
@@ -1276,7 +1286,7 @@ sound_init(void)
     sound_start_source_listener();
 
     /* save data from sound_server_source */
-    fifo_init(&g_in_fifo, 100);
+    g_in_fifo = fifo_create(in_fifo_item_destructor);
 
     g_client_does_fdk_aac = 0;
     g_client_fdk_aac_index = 0;
@@ -1293,6 +1303,12 @@ sound_init(void)
     }
     list_clear(g_ack_time_diff);
 
+#if defined(XRDP_FDK_AAC) || defined(XRDP_MP3LAME)
+    LOG(LOG_LEVEL_INFO, "num_silent_frames_aac: %d", g_cfg->num_silent_frames_aac);
+    LOG(LOG_LEVEL_INFO, "num_silent_frames_mp3: %d", g_cfg->num_silent_frames_mp3);
+    LOG(LOG_LEVEL_INFO, "msec_do_not_send: %d", g_cfg->msec_do_not_send);
+#endif
+
     return 0;
 }
 
@@ -1334,7 +1350,7 @@ sound_deinit(void)
     }
 #endif
 
-    fifo_deinit(&g_in_fifo);
+    fifo_delete(g_in_fifo, NULL);
 
     return 0;
 }
@@ -1654,7 +1670,6 @@ sound_process_input_formats(struct stream *s, int size)
 
     return 0;
 }
-
 /**
  *
  *****************************************************************************/
@@ -1667,10 +1682,7 @@ sound_input_start_recording(void)
     LOG_DEVEL(LOG_LEVEL_DEBUG, "sound_input_start_recording:");
 
     /* if there is any data in FIFO, discard it */
-    while ((s = (struct stream *) fifo_remove(&g_in_fifo)) != NULL)
-    {
-        xstream_free(s);
-    }
+    fifo_clear(g_in_fifo, NULL);
     g_bytes_in_fifo = 0;
 
     xstream_new(s, 1024);
@@ -1746,7 +1758,7 @@ sound_process_input_data(struct stream *s, int bytes)
     g_memcpy(ls->data, s->p, bytes);
     ls->p += bytes;
     s_mark_end(ls);
-    fifo_insert(&g_in_fifo, (void *) ls);
+    fifo_add_item(g_in_fifo, (void *) ls);
     g_bytes_in_fifo += bytes;
 
     return 0;
@@ -1799,7 +1811,7 @@ sound_sndsrvr_source_data_in(struct trans *trans)
         {
             if (g_stream_inp == NULL)
             {
-                g_stream_inp = (struct stream *) fifo_remove(&g_in_fifo);
+                g_stream_inp = (struct stream *) fifo_remove_item(g_in_fifo);
                 if (g_stream_inp != NULL)
                 {
                     g_bytes_in_fifo -= g_stream_inp->size;
@@ -1882,11 +1894,11 @@ sound_sndsrvr_source_data_in(struct trans *trans)
 static int
 sound_start_source_listener(void)
 {
-    char port[1024];
+    char port[XRDP_SOCKETS_MAXPATH];
 
     g_audio_l_trans_in = trans_create(TRANS_MODE_UNIX, 128 * 1024, 8192);
     g_audio_l_trans_in->is_term = g_is_term;
-    g_snprintf(port, 255, CHANSRV_PORT_IN_STR, g_display_num);
+    g_snprintf(port, sizeof(port), CHANSRV_PORT_IN_STR, g_getuid(), g_display_num);
     g_audio_l_trans_in->trans_conn_in = sound_sndsrvr_source_conn_in;
     if (trans_listen(g_audio_l_trans_in, port) != 0)
     {
@@ -1901,11 +1913,11 @@ sound_start_source_listener(void)
 static int
 sound_start_sink_listener(void)
 {
-    char port[1024];
+    char port[XRDP_SOCKETS_MAXPATH];
 
     g_audio_l_trans_out = trans_create(TRANS_MODE_UNIX, 128 * 1024, 8192);
     g_audio_l_trans_out->is_term = g_is_term;
-    g_snprintf(port, 255, CHANSRV_PORT_OUT_STR, g_display_num);
+    g_snprintf(port, sizeof(port), CHANSRV_PORT_OUT_STR, g_getuid(), g_display_num);
     g_audio_l_trans_out->trans_conn_in = sound_sndsrvr_sink_conn_in;
     if (trans_listen(g_audio_l_trans_out, port) != 0)
     {
