@@ -3,6 +3,7 @@
 #include "config_ac.h"
 #endif
 
+#include <stdio.h>
 #include <stdlib.h>
 
 #include "log.h"
@@ -45,11 +46,15 @@ int main (void)
     int number_failed;
     SRunner *sr;
 
-    sr = srunner_create (make_suite_test_list());
+    sr = srunner_create (make_suite_test_fifo());
+    srunner_add_suite(sr, make_suite_test_list());
+    srunner_add_suite(sr, make_suite_test_parse());
     srunner_add_suite(sr, make_suite_test_string());
+    srunner_add_suite(sr, make_suite_test_string_unicode());
     srunner_add_suite(sr, make_suite_test_os_calls());
     srunner_add_suite(sr, make_suite_test_ssl_calls());
     srunner_add_suite(sr, make_suite_test_base64());
+    srunner_add_suite(sr, make_suite_test_guid());
 
     srunner_set_tap(sr, "-");
     /*
@@ -57,15 +62,21 @@ int main (void)
     struct log_config *lc = log_config_init_for_console(LOG_LEVEL_INFO, NULL);
     log_start_from_param(lc);
     log_config_free(lc);
+    /* Disable stdout buffering, as this can confuse the error
+     * reporting when running in libcheck fork mode */
+    setvbuf(stdout, NULL, _IONBF, 0);
 
-    /* Initialise the ssl module */
+    /* Initialise modules */
     ssl_init();
+    os_calls_signals_init();
 
     srunner_run_all (sr, CK_ENV);
     number_failed = srunner_ntests_failed(sr);
     srunner_free(sr);
 
     ssl_finish();
+    os_calls_signals_deinit();
+
     log_end();
     return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
