@@ -4,5 +4,7 @@
 #include <check.h>
 
 Suite *make_suite_test_bitmap_load(void);
+Suite *make_suite_egfx_base_functions(void);
+Suite *make_suite_region(void);
 
 #endif /* TEST_XRDP_H */
