@@ -72,20 +72,32 @@ xrdp_wm_login_help_notify(struct xrdp_bitmap *wnd,
 
         if (p != 0)
         {
+            const int x = 10;
+            int y = xrdp_painter_font_body_height(p) * 2;
+            const int row_height = xrdp_painter_font_body_height(p);
+            const int end_para_height = row_height * 3 / 2;
+
             p->fg_color = wnd->wm->black;
-            xrdp_painter_draw_text(p, wnd, 10, 30, "You must be authenticated \
+            xrdp_painter_draw_text(p, wnd, x, y, "You must be authenticated \
 before using this");
-            xrdp_painter_draw_text(p, wnd, 10, 46, "session.");
-            xrdp_painter_draw_text(p, wnd, 10, 78, "Enter a valid username in \
+            y += row_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "session.");
+            y += end_para_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "Enter a valid username in \
 the username edit box.");
-            xrdp_painter_draw_text(p, wnd, 10, 94, "Enter the password in \
+            y += end_para_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "Enter the password in \
 the password edit box.");
-            xrdp_painter_draw_text(p, wnd, 10, 110, "Both the username and \
+            y += end_para_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "Both the username and \
 password are case");
-            xrdp_painter_draw_text(p, wnd, 10, 126, "sensitive.");
-            xrdp_painter_draw_text(p, wnd, 10, 158, "Contact your system \
+            y += row_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "sensitive.");
+            y += end_para_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "Contact your system \
 administrator if you are");
-            xrdp_painter_draw_text(p, wnd, 10, 174, "having problems \
+            y += row_height;
+            xrdp_painter_draw_text(p, wnd, x, y, "having problems \
 logging on.");
         }
     }
@@ -147,9 +159,23 @@ xrdp_wm_help_clicked(struct xrdp_bitmap *wnd)
 {
     struct xrdp_bitmap *help;
     struct xrdp_bitmap *but;
+    const int width =
+        wnd->wm->xrdp_config->cfg_globals.ls_scaled.help_wnd_width;
+    const int height =
+        wnd->wm->xrdp_config->cfg_globals.ls_scaled.help_wnd_height;
+    const int ok_height =
+        wnd->wm->xrdp_config->cfg_globals.ls_scaled.default_btn_height;
+    const char *ok_string = "OK";
+
+    /* Get a width for the OK button */
+    struct xrdp_painter *p = xrdp_painter_create(wnd->wm, wnd->wm->session);
+    xrdp_painter_font_needed(p);
+    const int ok_width = xrdp_painter_text_width(p, ok_string) +
+                         DEFAULT_BUTTON_MARGIN_W;
+    xrdp_painter_delete(p);
 
     /* create help screen */
-    help = xrdp_bitmap_create(DEFAULT_WND_HELP_W, DEFAULT_WND_HELP_H, wnd->wm->screen->bpp,
+    help = xrdp_bitmap_create(width, height, wnd->wm->screen->bpp,
                               WND_TYPE_WND, wnd->wm);
     list_insert_item(wnd->wm->screen->child_list, 0, (long)help);
     help->parent = wnd->wm->screen;
@@ -161,16 +187,16 @@ xrdp_wm_help_clicked(struct xrdp_bitmap *wnd)
     help->notify = xrdp_wm_login_help_notify;
     set_string(&help->caption1, "Login help");
     /* ok button */
-    but = xrdp_bitmap_create(DEFAULT_BUTTON_W, DEFAULT_BUTTON_H, wnd->wm->screen->bpp,
+    but = xrdp_bitmap_create(ok_width, ok_height, wnd->wm->screen->bpp,
                              WND_TYPE_BUTTON, wnd->wm);
     list_insert_item(help->child_list, 0, (long)but);
     but->parent = help;
     but->owner = help;
-    but->left = ((DEFAULT_WND_HELP_W / 2) - (DEFAULT_BUTTON_W / 2)); /* center */
-    but->top = DEFAULT_WND_HELP_H - DEFAULT_BUTTON_H - 15;
+    but->left = ((help->width / 2) - (ok_width / 2)); /* center */
+    but->top = help->height - ok_height - 15;
     but->id = 1;
     but->tab_stop = 1;
-    set_string(&but->caption1, "OK");
+    set_string(&but->caption1, ok_string);
     /* draw it */
     help->focused_control = but;
     help->default_button = but;
@@ -277,8 +303,7 @@ xrdp_wm_ok_clicked(struct xrdp_bitmap *wnd)
 */
 static int
 xrdp_wm_parse_domain_information(char *originalDomainInfo, int comboMax,
-                                 int decode,
-                                 char *resultBuffer, unsigned int resultSize)
+                                 int decode, char *resultBuffer)
 {
     int ret;
     int pos;
@@ -288,7 +313,8 @@ xrdp_wm_parse_domain_information(char *originalDomainInfo, int comboMax,
     /* If the first char in the domain name is '_' we use the domain
        name as IP*/
     ret = 0; /* default return value */
-    g_memset(resultBuffer, 0, resultSize);
+    /* resultBuffer assumed to be 256 chars */
+    g_memset(resultBuffer, 0, 256);
     if (originalDomainInfo[0] == '_')
     {
         /* we try to locate a number indicating what combobox index the user
@@ -298,7 +324,7 @@ xrdp_wm_parse_domain_information(char *originalDomainInfo, int comboMax,
          * Invalid chars are ignored in microsoft client therefore we use '_'
          * again. this sec '__' contains the split for index.*/
         pos = g_pos(&originalDomainInfo[1], "__");
-        if (pos > 0 && (unsigned int)pos < resultSize)
+        if (pos > 0)
         {
             /* an index is found we try to use it */
             LOG(LOG_LEVEL_DEBUG, "domain contains index char __");
@@ -325,7 +351,7 @@ xrdp_wm_parse_domain_information(char *originalDomainInfo, int comboMax,
         else
         {
             LOG(LOG_LEVEL_DEBUG, "domain does not contain _");
-            g_strncpy(resultBuffer, &originalDomainInfo[1], resultSize - 1);
+            g_strncpy(resultBuffer, &originalDomainInfo[1], 255);
         }
     }
     return ret;
@@ -387,32 +413,40 @@ xrdp_wm_show_edits(struct xrdp_wm *self, struct xrdp_bitmap *combo)
             }
             else if (g_strncmp(ASK, value, ASK_LEN) == 0)
             {
+                const int combo_height =
+                    self->xrdp_config->cfg_globals.ls_scaled.combo_height;
+                const int edit_height =
+                    self->xrdp_config->cfg_globals.ls_scaled.edit_height;
                 /* label */
-                b = xrdp_bitmap_create(globals->ls_label_width, DEFAULT_EDIT_H, self->screen->bpp,
+                b = xrdp_bitmap_create(globals->ls_scaled.label_width,
+                                       edit_height, self->screen->bpp,
                                        WND_TYPE_LABEL, self);
                 list_insert_item(self->login_window->child_list, insert_index,
                                  (long)b);
                 insert_index++;
                 b->parent = self->login_window;
                 b->owner = self->login_window;
-                b->left = globals->ls_label_x_pos;
+                b->left = globals->ls_scaled.label_x_pos;
 
-                b->top = globals->ls_input_y_pos + DEFAULT_COMBO_H + 5 + (DEFAULT_EDIT_H + 5) * count;
+                b->top = globals->ls_scaled.input_y_pos + combo_height + 5 +
+                         (edit_height + 5) * count;
                 b->id = 100 + 2 * count;
                 name = (char *)list_get_item(mod->names, index);
                 set_string(&b->caption1, name);
 
                 /* edit */
-                b = xrdp_bitmap_create(globals->ls_input_width, DEFAULT_EDIT_H, self->screen->bpp,
+                b = xrdp_bitmap_create(globals->ls_scaled.input_width,
+                                       edit_height, self->screen->bpp,
                                        WND_TYPE_EDIT, self);
                 list_insert_item(self->login_window->child_list, insert_index,
                                  (long)b);
                 insert_index++;
                 b->parent = self->login_window;
                 b->owner = self->login_window;
-                b->left = globals->ls_input_x_pos;
+                b->left = globals->ls_scaled.input_x_pos;
 
-                b->top = globals->ls_input_y_pos + DEFAULT_COMBO_H + 5 + (DEFAULT_EDIT_H + 5) * count;
+                b->top = globals->ls_scaled.input_y_pos + combo_height + 5 +
+                         (edit_height + 5) * count;
 
                 b->id = 100 + 2 * count + 1;
                 b->pointer = 1;
@@ -433,7 +467,7 @@ xrdp_wm_show_edits(struct xrdp_wm *self, struct xrdp_bitmap *combo)
                 {
                     g_strncpy(b->caption1, value + ASK_LEN, 255);
                 }
-                b->edit_pos = g_mbstowcs(0, b->caption1, 0);
+                b->edit_pos = utf8_char_count(b->caption1);
 
                 if (self->login_window->focused_control == 0)
                 {
@@ -450,10 +484,9 @@ xrdp_wm_show_edits(struct xrdp_wm *self, struct xrdp_bitmap *combo)
                     {
                         xrdp_wm_parse_domain_information(
                             self->session->client_info->domain,
-                            combo->data_list->count, 0,
-                            resultIP, sizeof(resultIP));
+                            combo->data_list->count, 0, resultIP);
                         g_strncpy(b->caption1, resultIP, 255);
-                        b->edit_pos = g_mbstowcs(0, b->caption1, 0);
+                        b->edit_pos = utf8_char_count(b->caption1);
                     }
 
                 }
@@ -462,7 +495,7 @@ xrdp_wm_show_edits(struct xrdp_wm *self, struct xrdp_bitmap *combo)
                         self->session->client_info->username[0])
                 {
                     g_strncpy(b->caption1, self->session->client_info->username, 255);
-                    b->edit_pos = g_mbstowcs(0, b->caption1, 0);
+                    b->edit_pos = utf8_char_count(b->caption1);
 
                     if (b->edit_pos > 0)
                     {
@@ -574,7 +607,7 @@ xrdp_wm_login_fill_in_combo(struct xrdp_wm *self, struct xrdp_bitmap *b)
     section_names->auto_free = 1;
     section_values = list_create();
     section_values->auto_free = 1;
-    fd = g_file_open(xrdp_ini);
+    fd = g_file_open_ro(xrdp_ini);
 
     if (fd < 0)
     {
@@ -619,11 +652,11 @@ xrdp_wm_login_fill_in_combo(struct xrdp_wm *self, struct xrdp_bitmap *b)
                     g_strncpy(name, r, 255);
                 }
 
-                list_add_item(mod_data->names, (long)g_strdup(q));
-                list_add_item(mod_data->values, (long)g_strdup(r));
+                list_add_strdup(mod_data->names, q);
+                list_add_strdup(mod_data->values, r);
             }
 
-            list_add_item(b->string_list, (long)g_strdup(name));
+            list_add_strdup(b->string_list, name);
             list_add_item(b->data_list, (long)mod_data);
         }
     }
@@ -635,6 +668,81 @@ xrdp_wm_login_fill_in_combo(struct xrdp_wm *self, struct xrdp_bitmap *b)
     return 0;
 }
 
+/******************************************************************************/
+unsigned int
+xrdp_login_wnd_get_monitor_dpi(struct xrdp_wm *self)
+{
+    unsigned int result = 0;
+    const struct display_size_description *display_sizes =
+            &self->client_info->display_sizes;
+    unsigned int height_pixels = 0;
+    unsigned int height_mm = 0;
+
+    unsigned int i;
+
+    /* Look at the monitor data first */
+    for (i = 0; i < display_sizes->monitorCount; ++i)
+    {
+        const struct monitor_info *mi = &display_sizes->minfo_wm[i];
+        {
+            if (mi->is_primary)
+            {
+                height_pixels = mi->bottom - mi->top + 1;
+                height_mm = mi->physical_height;
+                break;
+            }
+        }
+    }
+
+    /* No primary monitor, or values not defined - use the desktop size */
+    if (height_mm == 0)
+    {
+        height_pixels = display_sizes->session_height;
+        height_mm = self->client_info->session_physical_height;
+
+        if (height_mm == 0)
+        {
+            LOG(LOG_LEVEL_WARNING,
+                "No information is available to determine login screen DPI");
+        }
+        else if (height_pixels < 768)
+        {
+            /* A bug was encountered with mstsc.exe version
+               10.0.19041.1682 where the full physical monitor size was
+               sent in TS_UD_CS_CORE when the desktop size was set to
+               less than the screen size.
+               To generate the bug, make a connection with a full-screen
+               single window, cancel the login, and reconnect at
+               (e.g.) 800x600.
+               We can't detect that exact situation here, but if the
+               session height is so small as to likely be in a window
+               (rather than full screen), we should ignore the physical
+               size */
+            LOG(LOG_LEVEL_WARNING,
+                "Ignoring unlikely physical session size %u "
+                "for height of %u pixels", height_mm, height_pixels);
+            height_mm = 0;
+        }
+    }
+
+    if (height_mm != 0)
+    {
+        /*
+         * DPI = height_pixels / (height_mm / 25.4)
+         *     = (height_pixels * 25.4) / height_mm
+         *     = (height_pixels * 127) / (height_mm * 5)
+         */
+        result = (height_pixels * 127 ) / (height_mm * 5);
+        LOG(LOG_LEVEL_INFO,
+            "Login screen monitor height is %u pixels over %u mm (%u DPI)",
+            height_pixels,
+            height_mm,
+            result);
+    }
+    return result;
+}
+
+
 /******************************************************************************/
 int
 xrdp_login_wnd_create(struct xrdp_wm *self)
@@ -653,11 +761,15 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
     int primary_height;
     int primary_x_offset;  /* Offset of centre of primary screen */
     int primary_y_offset;
-    int index;
+    uint32_t index;
     int x;
     int y;
     int cx;
     int cy;
+    const int combo_height =
+        self->xrdp_config->cfg_globals.ls_scaled.combo_height;
+    const int edit_height =
+        self->xrdp_config->cfg_globals.ls_scaled.edit_height;
 
     globals = &self->xrdp_config->cfg_globals;
 
@@ -666,8 +778,8 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
     primary_x_offset = primary_width / 2;
     primary_y_offset = primary_height / 2;
 
-    log_width = globals->ls_width;
-    log_height = globals->ls_height;
+    log_width = globals->ls_scaled.width;
+    log_height = globals->ls_scaled.height;
     regular = 1;
 
     if (self->screen->width < log_width)
@@ -685,16 +797,16 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
     }
 
     /* multimon scenario, draw login window on primary monitor */
-    if (self->client_info->monitorCount > 1)
+    if (self->client_info->display_sizes.monitorCount > 1)
     {
-        for (index = 0; index < self->client_info->monitorCount; index++)
+        for (index = 0; index < self->client_info->display_sizes.monitorCount; index++)
         {
-            if (self->client_info->minfo_wm[index].is_primary)
+            if (self->client_info->display_sizes.minfo_wm[index].is_primary)
             {
-                x = self->client_info->minfo_wm[index].left;
-                y = self->client_info->minfo_wm[index].top;
-                cx = self->client_info->minfo_wm[index].right;
-                cy = self->client_info->minfo_wm[index].bottom;
+                x = self->client_info->display_sizes.minfo_wm[index].left;
+                y = self->client_info->display_sizes.minfo_wm[index].top;
+                cx = self->client_info->display_sizes.minfo_wm[index].right;
+                cy = self->client_info->display_sizes.minfo_wm[index].bottom;
 
                 primary_width = cx - x;
                 primary_height = cy - y;
@@ -798,7 +910,13 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
         /* if logo image not specified, use default */
         if (globals->ls_logo_filename[0] == 0)
         {
-            g_snprintf(globals->ls_logo_filename, 255, "%s/xrdp_logo.bmp", XRDP_SHARE_PATH);
+#ifdef USE_IMLIB2
+            g_snprintf(globals->ls_logo_filename, 255, "%s/xrdp_logo.png",
+                       XRDP_SHARE_PATH);
+#else
+            g_snprintf(globals->ls_logo_filename, 255, "%s/xrdp_logo.bmp",
+                       XRDP_SHARE_PATH);
+#endif
         }
 
         /* logo image */
@@ -809,61 +927,66 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
             g_snprintf(globals->ls_logo_filename, 255, "%s/ad256.bmp", XRDP_SHARE_PATH);
         }
 
+        LOG(LOG_LEVEL_DEBUG, "ls_logo_filename: %s", globals->ls_logo_filename);
+
         xrdp_bitmap_load(but, globals->ls_logo_filename, self->palette,
                          globals->ls_bg_color,
                          globals->ls_logo_transform,
-                         globals->ls_logo_width,
-                         globals->ls_logo_height);
+                         globals->ls_scaled.logo_width,
+                         globals->ls_scaled.logo_height);
         but->parent = self->login_window;
         but->owner = self->login_window;
-        but->left = globals->ls_logo_x_pos;
-        but->top = globals->ls_logo_y_pos;
+        but->left = globals->ls_scaled.logo_x_pos;
+        but->top = globals->ls_scaled.logo_y_pos;
         list_add_item(self->login_window->child_list, (long)but);
     }
 
     /* label */
-    but = xrdp_bitmap_create(globals->ls_label_width, DEFAULT_EDIT_H, self->screen->bpp, WND_TYPE_LABEL, self);
+    but = xrdp_bitmap_create(globals->ls_scaled.label_width, edit_height,
+                             self->screen->bpp, WND_TYPE_LABEL, self);
     list_add_item(self->login_window->child_list, (long)but);
     but->parent = self->login_window;
     but->owner = self->login_window;
-    but->left = globals->ls_label_x_pos;
-    but->top = globals->ls_input_y_pos;
+    but->left = globals->ls_scaled.label_x_pos;
+    but->top = globals->ls_scaled.input_y_pos;
     set_string(&but->caption1, "Session");
 
     /* combo */
-    combo = xrdp_bitmap_create(globals->ls_input_width, DEFAULT_COMBO_H,
+    combo = xrdp_bitmap_create(globals->ls_scaled.input_width, combo_height,
                                self->screen->bpp, WND_TYPE_COMBO, self);
     list_add_item(self->login_window->child_list, (long)combo);
     combo->parent = self->login_window;
     combo->owner = self->login_window;
-    combo->left = globals->ls_input_x_pos;
-    combo->top = globals->ls_input_y_pos;
+    combo->left = globals->ls_scaled.input_x_pos;
+    combo->top = globals->ls_scaled.input_y_pos;
     combo->id = 6;
     combo->tab_stop = 1;
     xrdp_wm_login_fill_in_combo(self, combo);
 
     /* OK button */
-    but = xrdp_bitmap_create(globals->ls_btn_ok_width, globals->ls_btn_ok_height,
+    but = xrdp_bitmap_create(globals->ls_scaled.btn_ok_width,
+                             globals->ls_scaled.btn_ok_height,
                              self->screen->bpp, WND_TYPE_BUTTON, self);
     list_add_item(self->login_window->child_list, (long)but);
     but->parent = self->login_window;
     but->owner = self->login_window;
-    but->left = globals->ls_btn_ok_x_pos;
-    but->top = globals->ls_btn_ok_y_pos;
+    but->left = globals->ls_scaled.btn_ok_x_pos;
+    but->top = globals->ls_scaled.btn_ok_y_pos;
     but->id = 3;
     set_string(&but->caption1, "OK");
     but->tab_stop = 1;
     self->login_window->default_button = but;
 
     /* Cancel button */
-    but = xrdp_bitmap_create(globals->ls_btn_cancel_width,
-                             globals->ls_btn_cancel_height, self->screen->bpp,
+    but = xrdp_bitmap_create(globals->ls_scaled.btn_cancel_width,
+                             globals->ls_scaled.btn_cancel_height,
+                             self->screen->bpp,
                              WND_TYPE_BUTTON, self);
     list_add_item(self->login_window->child_list, (long)but);
     but->parent = self->login_window;
     but->owner = self->login_window;
-    but->left = globals->ls_btn_cancel_x_pos;
-    but->top = globals->ls_btn_cancel_y_pos;
+    but->left = globals->ls_scaled.btn_cancel_x_pos;
+    but->top = globals->ls_scaled.btn_cancel_y_pos;
     but->id = 2;
     set_string(&but->caption1, "Cancel");
     but->tab_stop = 1;
@@ -876,8 +999,7 @@ xrdp_login_wnd_create(struct xrdp_wm *self)
     combo->item_index = xrdp_wm_parse_domain_information(
                             self->session->client_info->domain,
                             combo->data_list->count, 1,
-                            resultIP,/* just a dummy place holder, we ignore */
-                            sizeof(resultIP));
+                            resultIP /* just a dummy place holder, we ignore */ );
     xrdp_wm_show_edits(self, combo);
 
     return 0;
@@ -950,30 +1072,42 @@ load_xrdp_config(struct xrdp_config *config, const char *xrdp_ini, int bpp)
 
     /* set default values in case we can't get them from xrdp.ini file */
     globals->ini_version = 1;
+    globals->default_dpi = 96;
+
     globals->ls_top_window_bg_color = HCOLOR(bpp, xrdp_wm_htoi("009cb5"));
     globals->ls_bg_color = HCOLOR(bpp, xrdp_wm_htoi("dedede"));
-    globals->ls_width = 350;
-    globals->ls_height = 350;
+    globals->ls_unscaled.width = 350;
+    globals->ls_unscaled.height = 350;
     globals->ls_background_transform = XBLT_NONE;
     globals->ls_logo_transform = XBLT_NONE;
-    globals->ls_logo_x_pos = 63;
-    globals->ls_logo_y_pos = 50;
-    globals->ls_label_x_pos = 30;
-    globals->ls_label_width = 65;
-    globals->ls_input_x_pos = 110;
-    globals->ls_input_width = 210;
-    globals->ls_input_y_pos = 150;
-    globals->ls_btn_ok_x_pos = 150;
-    globals->ls_btn_ok_y_pos = 300;
-    globals->ls_btn_ok_width = 85;
-    globals->ls_btn_ok_height = 30;
-    globals->ls_btn_cancel_x_pos = 245;
-    globals->ls_btn_cancel_y_pos = 300;
-    globals->ls_btn_cancel_width = 85;
-    globals->ls_btn_cancel_height = 30;
+    globals->ls_unscaled.logo_x_pos = 63;
+    globals->ls_unscaled.logo_y_pos = 50;
+    globals->ls_unscaled.label_x_pos = 30;
+    globals->ls_unscaled.label_width = 65;
+    globals->ls_unscaled.input_x_pos = 110;
+    globals->ls_unscaled.input_width = 210;
+    globals->ls_unscaled.input_y_pos = 150;
+    globals->ls_unscaled.btn_ok_x_pos = 150;
+    globals->ls_unscaled.btn_ok_y_pos = 300;
+    globals->ls_unscaled.btn_ok_width = 85;
+    globals->ls_unscaled.btn_ok_height = 30;
+    globals->ls_unscaled.btn_cancel_x_pos = 245;
+    globals->ls_unscaled.btn_cancel_y_pos = 300;
+    globals->ls_unscaled.btn_cancel_width = 85;
+    globals->ls_unscaled.btn_cancel_height = 30;
+    globals->ls_unscaled.default_btn_height =
+        DEFAULT_FONT_PIXEL_SIZE + DEFAULT_BUTTON_MARGIN_H;
+    globals->ls_unscaled.log_wnd_width = DEFAULT_WND_LOG_W;
+    globals->ls_unscaled.log_wnd_height = DEFAULT_WND_LOG_H;
+    globals->ls_unscaled.edit_height =
+        DEFAULT_FONT_PIXEL_SIZE + DEFAULT_EDIT_MARGIN_H;
+    globals->ls_unscaled.combo_height =
+        DEFAULT_FONT_PIXEL_SIZE + DEFAULT_COMBO_MARGIN_H;
+    globals->ls_unscaled.help_wnd_width = DEFAULT_WND_HELP_W;
+    globals->ls_unscaled.help_wnd_height = DEFAULT_WND_HELP_H;
 
     /* open xrdp.ini file */
-    if ((fd = g_file_open(xrdp_ini)) < 0)
+    if ((fd = g_file_open_ro(xrdp_ini)) < 0)
     {
         LOG(LOG_LEVEL_ERROR, "load_config: Could not read "
             "xrdp.ini file %s", xrdp_ini);
@@ -1167,6 +1301,16 @@ load_xrdp_config(struct xrdp_config *config, const char *xrdp_ini, int bpp)
         }
 
         /* login screen values */
+        else if (g_strcmp(n, "default_dpi") == 0)
+        {
+            globals->default_dpi = g_atoi(v);
+        }
+
+        else if (g_strcmp(n, "fv1_select") == 0)
+        {
+            g_strncpy(globals->fv1_select, v, sizeof(globals->fv1_select) - 1);
+        }
+
         else if (g_strncmp(n, "ls_top_window_bg_color", 64) == 0)
         {
             globals->ls_top_window_bg_color = HCOLOR(bpp, xrdp_wm_htoi(v));
@@ -1174,12 +1318,12 @@ load_xrdp_config(struct xrdp_config *config, const char *xrdp_ini, int bpp)
 
         else if (g_strncmp(n, "ls_width", 64) == 0)
         {
-            globals->ls_width = g_atoi(v);
+            globals->ls_unscaled.width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_height", 64) == 0)
         {
-            globals->ls_height = g_atoi(v);
+            globals->ls_unscaled.height = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_bg_color", 64) == 0)
@@ -1218,87 +1362,87 @@ load_xrdp_config(struct xrdp_config *config, const char *xrdp_ini, int bpp)
 
         else if (g_strncmp(n, "ls_logo_width", 64) == 0)
         {
-            globals->ls_logo_width = g_atoi(v);
+            globals->ls_unscaled.logo_width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_logo_height", 64) == 0)
         {
-            globals->ls_logo_height = g_atoi(v);
+            globals->ls_unscaled.logo_height = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_logo_x_pos", 64) == 0)
         {
-            globals->ls_logo_x_pos = g_atoi(v);
+            globals->ls_unscaled.logo_x_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_logo_y_pos", 64) == 0)
         {
-            globals->ls_logo_y_pos = g_atoi(v);
+            globals->ls_unscaled.logo_y_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_label_x_pos", 64) == 0)
         {
-            globals->ls_label_x_pos = g_atoi(v);
+            globals->ls_unscaled.label_x_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_label_width", 64) == 0)
         {
-            globals->ls_label_width = g_atoi(v);
+            globals->ls_unscaled.label_width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_input_x_pos", 64) == 0)
         {
-            globals->ls_input_x_pos = g_atoi(v);
+            globals->ls_unscaled.input_x_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_input_width", 64) == 0)
         {
-            globals->ls_input_width = g_atoi(v);
+            globals->ls_unscaled.input_width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_input_y_pos", 64) == 0)
         {
-            globals->ls_input_y_pos = g_atoi(v);
+            globals->ls_unscaled.input_y_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_ok_x_pos", 64) == 0)
         {
-            globals->ls_btn_ok_x_pos = g_atoi(v);
+            globals->ls_unscaled.btn_ok_x_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_ok_y_pos", 64) == 0)
         {
-            globals->ls_btn_ok_y_pos = g_atoi(v);
+            globals->ls_unscaled.btn_ok_y_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_ok_width", 64) == 0)
         {
-            globals->ls_btn_ok_width = g_atoi(v);
+            globals->ls_unscaled.btn_ok_width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_ok_height", 64) == 0)
         {
-            globals->ls_btn_ok_height = g_atoi(v);
+            globals->ls_unscaled.btn_ok_height = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_cancel_x_pos", 64) == 0)
         {
-            globals->ls_btn_cancel_x_pos = g_atoi(v);
+            globals->ls_unscaled.btn_cancel_x_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_cancel_y_pos", 64) == 0)
         {
-            globals->ls_btn_cancel_y_pos = g_atoi(v);
+            globals->ls_unscaled.btn_cancel_y_pos = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_cancel_width", 64) == 0)
         {
-            globals->ls_btn_cancel_width = g_atoi(v);
+            globals->ls_unscaled.btn_cancel_width = g_atoi(v);
         }
 
         else if (g_strncmp(n, "ls_btn_cancel_height", 64) == 0)
         {
-            globals->ls_btn_cancel_height = g_atoi(v);
+            globals->ls_unscaled.btn_cancel_height = g_atoi(v);
         }
     }
 
@@ -1335,29 +1479,95 @@ load_xrdp_config(struct xrdp_config *config, const char *xrdp_ini, int bpp)
     LOG(LOG_LEVEL_DEBUG, "enable_token_login:      %d", globals->enable_token_login);
 
     LOG(LOG_LEVEL_DEBUG, "ls_top_window_bg_color:  %x", globals->ls_top_window_bg_color);
-    LOG(LOG_LEVEL_DEBUG, "ls_width:                %d", globals->ls_width);
-    LOG(LOG_LEVEL_DEBUG, "ls_height:               %d", globals->ls_height);
+    LOG(LOG_LEVEL_DEBUG, "ls_width (unscaled):     %d", globals->ls_unscaled.width);
+    LOG(LOG_LEVEL_DEBUG, "ls_height (unscaled):    %d", globals->ls_unscaled.height);
     LOG(LOG_LEVEL_DEBUG, "ls_bg_color:             %x", globals->ls_bg_color);
     LOG(LOG_LEVEL_DEBUG, "ls_title:                %s", globals->ls_title);
     LOG(LOG_LEVEL_DEBUG, "ls_logo_filename:        %s", globals->ls_logo_filename);
-    LOG(LOG_LEVEL_DEBUG, "ls_logo_x_pos:           %d", globals->ls_logo_x_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_logo_y_pos:           %d", globals->ls_logo_y_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_label_x_pos:          %d", globals->ls_label_x_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_label_width:          %d", globals->ls_label_width);
-    LOG(LOG_LEVEL_DEBUG, "ls_input_x_pos:          %d", globals->ls_input_x_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_input_width:          %d", globals->ls_input_width);
-    LOG(LOG_LEVEL_DEBUG, "ls_input_y_pos:          %d", globals->ls_input_y_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_x_pos:         %d", globals->ls_btn_ok_x_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_y_pos:         %d", globals->ls_btn_ok_y_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_width:         %d", globals->ls_btn_ok_width);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_height:        %d", globals->ls_btn_ok_height);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_x_pos:     %d", globals->ls_btn_cancel_x_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_y_pos:     %d", globals->ls_btn_cancel_y_pos);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_width:     %d", globals->ls_btn_cancel_width);
-    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_height:    %d", globals->ls_btn_cancel_height);
+    LOG(LOG_LEVEL_DEBUG, "ls_logo_x_pos :          %d", globals->ls_unscaled.logo_x_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_logo_y_pos :          %d", globals->ls_unscaled.logo_y_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_label_x_pos :         %d", globals->ls_unscaled.label_x_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_label_width :         %d", globals->ls_unscaled.label_width);
+    LOG(LOG_LEVEL_DEBUG, "ls_input_x_pos :         %d", globals->ls_unscaled.input_x_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_input_width :         %d", globals->ls_unscaled.input_width);
+    LOG(LOG_LEVEL_DEBUG, "ls_input_y_pos :         %d", globals->ls_unscaled.input_y_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_x_pos :        %d", globals->ls_unscaled.btn_ok_x_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_y_pos :        %d", globals->ls_unscaled.btn_ok_y_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_width :        %d", globals->ls_unscaled.btn_ok_width);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_ok_height :       %d", globals->ls_unscaled.btn_ok_height);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_x_pos :    %d", globals->ls_unscaled.btn_cancel_x_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_y_pos :    %d", globals->ls_unscaled.btn_cancel_y_pos);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_width :    %d", globals->ls_unscaled.btn_cancel_width);
+    LOG(LOG_LEVEL_DEBUG, "ls_btn_cancel_height :   %d", globals->ls_unscaled.btn_cancel_height);
 
     list_delete(names);
     list_delete(values);
     g_file_close(fd);
     return 0;
 }
+
+/**
+ * Scale the configuration values
+ *
+ * After a font has been loaded, we can produce scaled versions of the
+ * login screen layout parameters which will correspond to the size of the
+ * font
+ */
+void
+xrdp_login_wnd_scale_config_values(struct xrdp_wm *self)
+{
+    const struct xrdp_ls_dimensions *unscaled =
+            &self->xrdp_config->cfg_globals.ls_unscaled;
+    struct xrdp_ls_dimensions *scaled =
+            &self->xrdp_config->cfg_globals.ls_scaled;
+
+    /* Clear the scaled values, so if we add one and forget to scale it,
+     * it will be obvious */
+    g_memset(scaled, '\0', sizeof(*scaled));
+
+    /* If we don't have a font, use zeros for everything */
+    if (self->default_font == NULL)
+    {
+        LOG(LOG_LEVEL_ERROR, "Can't scale login values - no font available");
+    }
+    else
+    {
+        const int fheight = self->default_font->body_height;
+        /* Define a Macro to scale to the nearest pixel value,
+         * rounding up as appropriate */
+#define SCALE_AND_ROUND(x) \
+    (((x) * fheight + (DEFAULT_FONT_PIXEL_SIZE / 2)) / \
+     DEFAULT_FONT_PIXEL_SIZE)
+
+        LOG(LOG_LEVEL_DEBUG, "Login screen scale factor %f",
+            (float)fheight / DEFAULT_FONT_PIXEL_SIZE);
+
+        scaled->width = SCALE_AND_ROUND(unscaled->width);
+        scaled->height = SCALE_AND_ROUND(unscaled->height);
+        scaled->logo_width = SCALE_AND_ROUND(unscaled->logo_width);
+        scaled->logo_height = SCALE_AND_ROUND(unscaled->logo_height);
+        scaled->logo_x_pos = SCALE_AND_ROUND(unscaled->logo_x_pos);
+        scaled->logo_y_pos = SCALE_AND_ROUND(unscaled->logo_y_pos);
+        scaled->label_x_pos = SCALE_AND_ROUND(unscaled->label_x_pos);
+        scaled->label_width = SCALE_AND_ROUND(unscaled->label_width);
+        scaled->input_x_pos = SCALE_AND_ROUND(unscaled->input_x_pos);
+        scaled->input_width = SCALE_AND_ROUND(unscaled->input_width);
+        scaled->input_y_pos = SCALE_AND_ROUND(unscaled->input_y_pos);
+        scaled->btn_ok_x_pos = SCALE_AND_ROUND(unscaled->btn_ok_x_pos);
+        scaled->btn_ok_y_pos = SCALE_AND_ROUND(unscaled->btn_ok_y_pos);
+        scaled->btn_ok_width = SCALE_AND_ROUND(unscaled->btn_ok_width);
+        scaled->btn_ok_height = SCALE_AND_ROUND(unscaled->btn_ok_height);
+        scaled->btn_cancel_x_pos = SCALE_AND_ROUND(unscaled->btn_cancel_x_pos);
+        scaled->btn_cancel_y_pos = SCALE_AND_ROUND(unscaled->btn_cancel_y_pos);
+        scaled->btn_cancel_width = SCALE_AND_ROUND(unscaled->btn_cancel_width);
+        scaled->btn_cancel_height = SCALE_AND_ROUND(unscaled->btn_cancel_height);
+        scaled->default_btn_height = fheight + DEFAULT_BUTTON_MARGIN_H;
+        scaled->log_wnd_width = SCALE_AND_ROUND(unscaled->log_wnd_width);
+        scaled->log_wnd_height = SCALE_AND_ROUND(unscaled->log_wnd_height);
+        scaled->edit_height = fheight + DEFAULT_EDIT_MARGIN_H;
+        scaled->combo_height = fheight + DEFAULT_COMBO_MARGIN_H;
+        scaled->help_wnd_width = SCALE_AND_ROUND(unscaled->help_wnd_width);
+        scaled->help_wnd_height = SCALE_AND_ROUND(unscaled->help_wnd_height);
+#undef SCALE_AND_ROUND
+    }
+}
