@@ -32,6 +32,7 @@
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <poll.h>
 
 #include "log.h"
 #include "xrdp_sockets.h"
@@ -46,9 +47,9 @@ struct wts_obj
 
 /* helper functions used by WTSxxx API - do not invoke directly */
 static int
-can_send(int sck, int millis);
+can_send(int sck, int millis, int restart);
 static int
-can_recv(int sck, int millis);
+can_recv(int sck, int millis, int restart);
 static int
 mysend(int sck, const void *adata, int bytes);
 static int
@@ -139,7 +140,7 @@ WTSVirtualChannelOpenEx(unsigned int SessionId, const char *pVirtualName,
     memset(&s, 0, sizeof(struct sockaddr_un));
     s.sun_family = AF_UNIX;
     bytes = sizeof(s.sun_path);
-    snprintf(s.sun_path, bytes - 1, CHANSRV_API_STR, wts->display_num);
+    snprintf(s.sun_path, bytes - 1, CHANSRV_API_STR, getuid(), wts->display_num);
     s.sun_path[bytes - 1] = 0;
     bytes = sizeof(struct sockaddr_un);
 
@@ -160,7 +161,7 @@ WTSVirtualChannelOpenEx(unsigned int SessionId, const char *pVirtualName,
     }
 
     /* wait for connection to complete */
-    if (!can_send(wts->fd, 500))
+    if (!can_send(wts->fd, 500, 1))
     {
         LOG(LOG_LEVEL_ERROR, "WTSVirtualChannelOpenEx: can_send failed");
         free(wts);
@@ -212,7 +213,7 @@ WTSVirtualChannelOpenEx(unsigned int SessionId, const char *pVirtualName,
     }
     LOG_DEVEL(LOG_LEVEL_DEBUG, "WTSVirtualChannelOpenEx: sent ok");
 
-    if (!can_recv(wts->fd, 500))
+    if (!can_recv(wts->fd, 500, 1))
     {
         LOG(LOG_LEVEL_ERROR, "WTSVirtualChannelOpenEx: can_recv failed");
         free(wts);
@@ -263,7 +264,7 @@ mysend(int sck, const void *adata, int bytes)
     sent = 0;
     while (sent < bytes)
     {
-        if (can_send(sck, 100))
+        if (can_send(sck, 100, 0))
         {
             error = send(sck, data + sent, bytes - sent, MSG_NOSIGNAL);
             if (error < 1)
@@ -293,7 +294,7 @@ myrecv(int sck, void *adata, int bytes)
     recd = 0;
     while (recd < bytes)
     {
-        if (can_recv(sck, 100))
+        if (can_recv(sck, 100, 0))
         {
             error = recv(sck, data + recd, bytes - recd, MSG_NOSIGNAL);
             if (error < 1)
@@ -328,7 +329,7 @@ WTSVirtualChannelWrite(void *hChannelHandle, const char *Buffer,
         return 0;
     }
 
-    if (!can_send(wts->fd, 0))
+    if (!can_send(wts->fd, 0, 0))
     {
         return 1;    /* can't write now, ok to try again */
     }
@@ -369,7 +370,7 @@ WTSVirtualChannelRead(void *hChannelHandle, unsigned int TimeOut,
         return 0;
     }
 
-    if (can_recv(wts->fd, TimeOut))
+    if (can_recv(wts->fd, TimeOut, 0))
     {
         rv = recv(wts->fd, Buffer, BufferSize, 0);
 
@@ -474,47 +475,63 @@ WTSFreeMemory(void *pMemory)
  *
  * @param  sck    socket to check
  * @param  millis timeout value in milliseconds
+ * @param  restart Try again if interrupted, even if this exceeds the timeout
  *
  * @return 0 if write will block
  * @return 1 if write will not block
  ******************************************************************************/
 static int
-can_send(int sck, int millis)
+can_send(int sck, int millis, int restart)
 {
-    struct timeval time;
-    fd_set         wfds;
-    int            select_rv;
+    int rv = 0;
+    struct pollfd pollfd;
+    int status;
 
-    /* setup for a select call */
-    FD_ZERO(&wfds);
-    FD_SET(sck, &wfds);
-    time.tv_sec = millis / 1000;
-    time.tv_usec = (millis * 1000) % 1000000;
+    pollfd.fd = sck;
+    pollfd.events = POLLOUT;
+    pollfd.revents = 0;
 
-    /* check if it is ok to write to specified socket */
-    select_rv = select(sck + 1, 0, &wfds, 0, &time);
+    do
+    {
+        status = poll(&pollfd, 1, millis);
+    }
+    while (status < 0 && errno == EINTR && restart);
 
-    return (select_rv > 0) ? 1 : 0;
+    if (status > 0)
+    {
+        if ((pollfd.revents & POLLOUT) != 0)
+        {
+            rv = 1;
+        }
+    }
+
+    return rv;
 }
 
 /*****************************************************************************/
 static int
-can_recv(int sck, int millis)
+can_recv(int sck, int millis, int restart)
 {
-    struct timeval time;
-    fd_set rfds;
-    int select_rv;
+    int rv = 0;
+    struct pollfd pollfd;
+    int status;
 
-    FD_ZERO(&rfds);
-    FD_SET(sck, &rfds);
-    time.tv_sec = millis / 1000;
-    time.tv_usec = (millis * 1000) % 1000000;
-    select_rv = select(sck + 1, &rfds, 0, 0, &time);
-
-    if (select_rv > 0)
+    pollfd.fd = sck;
+    pollfd.events = POLLIN;
+    pollfd.revents = 0;
+    do
     {
-        return 1;
+        status = poll(&pollfd, 1, millis);
+    }
+    while (status < 0 && errno == EINTR && restart);
+
+    if (status > 0)
+    {
+        if ((pollfd.revents & (POLLIN | POLLHUP)) != 0)
+        {
+            rv = 1;
+        }
     }
 
-    return 0;
+    return rv;
 }
